__version__ = "0.5.3"
__copyright__    = 'Copyright (C) 2023 CHIGIRA Koki'
__license__      = 'MIT License'
__author__       = 'CHIGIRA Koki'
__author_email__ = 'kyoujin2009kutar@gmail.com'
__url__          = 'http://github.com/KChigira/mkdesigner'
